int CELL_SIZE = 4;
int PIXEL_WIDTH = 640;
int PIXEL_HEIGHT = 480;
float INITIAL_POPULATION_DENSITY = 0.5;

float[] COLOR_DIVISIONS = {0.0, 0.25, 0.5, 0.75, 1.0};
color[] COLORS = {color(0,0,0), color(14,0,117), color(0,86,191), color(0,207,251), color(255,255,255)};
int FRAME_RATE = 5;

int WORLD_WIDTH = PIXEL_WIDTH/CELL_SIZE;
int WORLD_HEIGHT = PIXEL_HEIGHT/CELL_SIZE;
int[][] world;
float[][] averages;


void setup() {
  size(PIXEL_WIDTH, PIXEL_HEIGHT);
  frameRate(FRAME_RATE);
  colorMode(RGB, 255);
  world = new int[WORLD_WIDTH][WORLD_HEIGHT];
  for (int i=0; i<WORLD_WIDTH * WORLD_HEIGHT * INITIAL_POPULATION_DENSITY; i++) {
    world[(int)random(WORLD_WIDTH)][(int)random(WORLD_HEIGHT)] = 1;
  }
  averages = new float[WORLD_WIDTH][WORLD_HEIGHT];
}

void draw() {
  background(0);
  evolve();
  average();
  draw_world();
}

void evolve() {
  int[][] new_world = new int[WORLD_WIDTH][WORLD_HEIGHT];
  for (int x=1; x<WORLD_WIDTH-1; x++) {
    for (int y=1; y<WORLD_HEIGHT-1; y++) {
      // add up the neighbers from top left, clockwise
      int neighbors = world[x-1][y-1] +
                      world[x][y-1] +
                      world[x+1][y-1] +
                      world[x-1][y] +
                      world[x+1][y] +
                      world[x-1][y+1] +
                      world[x][y+1] +
                      world[x+1][y+1];
      
      boolean empty = world[x][y] == 0;
      int alive = 0;
      if (!empty) {
        if (neighbors < 2) {
          alive = 0;
        } else if (neighbors <= 3) {
          alive = 1;
        } else {
          alive = 0;
        }
      } else {
        if (neighbors == 3) {
          alive = 1;
        }
      }
      new_world[x][y] = alive;
    }
  }
  arrayCopy(new_world, world);
}

void average() {
  float[][] new_averages = new float[WORLD_WIDTH][WORLD_HEIGHT];
  for (int x=1; x<WORLD_WIDTH-1; x++) {
    for (int y=1; y<WORLD_HEIGHT-1; y++) {
      new_averages[x][y] = averages[x][y] * (frameCount-1);
      new_averages[x][y] += world[x][y];
      new_averages[x][y] /= (float)frameCount;
    }
  }
  arrayCopy(new_averages, averages);
}

void draw_world() {
  noStroke();
  for (int x=0; x<WORLD_WIDTH; x++) {
    for (int y=0; y<WORLD_HEIGHT; y++) {
      fill(color_for_value(averages[x][y]));
      rect(x*CELL_SIZE, y*CELL_SIZE, CELL_SIZE, CELL_SIZE);
    }
  }
}

color color_for_value(float value) {
  for(int i=COLOR_DIVISIONS.length-2; i>=0; i--) {
    if(value > COLOR_DIVISIONS[i]) {
      float subtracted_value = value - COLOR_DIVISIONS[i];
      float normalized_value = subtracted_value / (COLOR_DIVISIONS[i+1] - COLOR_DIVISIONS[i]);
      color low_color = COLORS[i];
      color high_color = COLORS[i+1];
      color new_color = color(red(low_color) + (red(high_color) - red(low_color))*normalized_value,
        green(low_color) + (green(high_color) - green(low_color))*normalized_value,
        blue(low_color) + (blue(high_color) - blue(low_color))*normalized_value);
      return new_color;
    }
  }
  return color(0);
}
